! Author: Adelchi Azzalini, Universit\`a di Padova, Italia
! Dec.1998 (minor modification in Sept.2008)
!
program mle
  ! use prompting
  use sn
  use machine_constants
  use toms611
  USE symxinv
  ! USE winteracter
  implicit none
  integer :: i, j, n, p, p0, npar, inputs(3), nf, nullty, ifault
  integer :: iv(60), liv=60, lv=150
  INTEGER, DIMENSION(:), ALLOCATABLE :: b
  real(dp) :: v(150), ufparm, dev, det
  real(dp), dimension (:), allocatable :: param, step, se
  real(dp), dimension (:,:), allocatable :: X, Mat
  real(dp), dimension (:), allocatable :: y, Xy, vect1, vect2, vect
  character(len=60, kind=1) :: datafile

  INTERFACE
    subroutine prompt_int(message,value)
      implicit none
      character (LEN=*), INTENT(IN) ::message
      INTEGER, INTENT(OUT) :: value
    end subroutine prompt_int
    subroutine obj_f(npar, param, nf, dev, inputs, Xy, ufparm)
      use sn
      implicit none
      INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(15, 60)
      real(dp), intent(in) :: param(:)
      real(dp), intent(out) :: dev
      real(dp), intent(in out) ::  Xy(:), ufparm
      integer, intent(in) :: npar, inputs(:), nf  
      end subroutine obj_f
    subroutine obj_gh(npar, param, nf, vect1, vect2, inputs, Xy, ufparm)
      implicit none
      INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(15, 60)
      real(dp), intent(in) :: param(:)
      real(dp), intent(out) :: vect1(:), vect2(:)
      real(dp), intent(in out) ::  Xy(:), ufparm
      integer, intent(in) :: npar, nf, inputs(:)  
      end subroutine obj_gh
    !SUBROUTINE subinv(a, nm, b, n, c, nullty, ifault, det)
    !  IMPLICIT NONE
    !  INTEGER, PARAMETER               :: dp = SELECTED_REAL_KIND(15, 60)
    !  REAL (dp), INTENT(IN)            :: a(:)
    !  INTEGER, INTENT(IN)              :: nm, b(:), n
    !  REAL (dp), INTENT(OUT)           :: c(:)
    !  REAL (dp), INTENT(OUT), OPTIONAL :: det
    !  INTEGER, INTENT(OUT)             :: nullty, ifault
    !END SUBROUTINE subinv
  END INTERFACE

  write(*,*) "MLE for regression models with skew-normal errors"
  write(*,*) "Home page of this program is ", &
      "http://azzalini.stat.unipd.it/SN"
  WRITE(*,*) "see this page for background documentation"
  WRITE(*,*)
  write(*,'(a)', advance='no') "Data file: "
  READ(*,*) datafile
  call prompt_int("Number of records: ",n)
  call prompt_int("Number of covariates (exluding constant): ",p0)
  call prompt_int("Print details of iterations? (type 0 for NO, 1 for YES): ",i)
  open(unit=5, file=datafile, status="old", action="read")
  p=p0+1
  npar=p+2
  inputs(1)=n
  inputs(2)=p
  inputs(3)=i
  allocate (X(n,p), y(n), Xy(n*(p+1)), param(npar), step(npar),        &
      se(npar), vect1(npar) , vect2(npar*(npar+1)/2), b(npar),         &
      Mat(p,p), vect(n) )
  do i=1,npar
    step(i)=1.0
    param(i)=0.0
    b(i)=i
  end do
  do i=1, n
    X(i,1)=1.0_dp
    read (5,*) (X(i,j+1),j=1,p0), y(i)
  end do
  close(5)
  ! Xy=(/ reshape(X, (/ n*(p+1), 1/) ), y /)  ! Sun f90 Compiler CRASH !!
  Xy=(/X,y/)

!.. calcola stime iniziali di param(1:p) con i minimi quadrati,
!   e da queste ottiene quelle di sigma, gamma1
  Mat=MATMUL(TRANSPOSE(X),X)
  do i=1,p
    do j=1,i
      vect2(i*(i-1)/2+j)=Mat(i,j)
    end do
  end do
  call subinv(vect2,npar*(npar+1)/2,b,p,vect2,nullty,ifault,det)
  IF(nullty+ifault > 0)                                                 &
    write(*,*) "nullty, ifault, det: ",nullty,ifault,det
  do i=1,p
    do j=1,i
      Mat(i,j)=vect2(i*(i-1)/2+j)
      Mat(j,i)=Mat(i,j)
    end do
  end do
  param(1:p)= MATMUL(Mat,MATMUL(transpose(X),y))
  vect = y-MATMUL(X,param(1:p))
  param(p+1)=SQRT(SUM(vect**2)/n)
  param(p+2)=SUM(vect**3)/(n*param(p+1)**3)
  IF(ABS(param(p+2))>0.99) param(p+2)=SIGN(0.99_dp,param(p+2))

  call deflt(2,iv,liv,lv,v)
  iv(19)=0
  iv(21)=0
  call humsl(npar,step,param, obj_f,obj_gh, iv,liv,lv,v, inputs, Xy, ufparm)
!
!.. fine ottimizzazione: stampa risultati finali
  WRITE(*,*)
  select case (iv(1))
    case (3:6)
      write(*,*)"Convergence reached."
    case (7,8)
      write(*,*)"Singular or false convergence."
    case default
      write(*,*)"No convergence."
  end select
  inputs(3)=0
  nf=1  ! fake ELF90
  call obj_f(npar, param, nf, dev, inputs, Xy, ufparm)
  write(*,93) "max log(likelihood) ", -dev/two
  call obj_gh(npar, param, nf, vect1, vect2, inputs, Xy, ufparm)
  write(*,93) "Norm of gradient: ", SUM(vect1**2)
  write(*,93) "MLEs of 'centred' parameters (beta(1:p),sigma,gamma1): "
  write(*,90) param
  write(*,93) "Observed Fisher information matrix:"
  do i=1,npar
    write(*,90) vect2(i*(i-1)/2+1:(i+1)*i/2)/two
  end do
  call subinv(vect2,npar*(npar+1)/2,b,npar,vect2,nullty,ifault,det)
  IF(nullty+ifault > 0)                                                 &
    write(*,*)"nullty, ifault, nf, det: ", nullty, ifault, nf, det
  do i=1,npar
    se(i)=SQRT(two * vect2(i*(i+1)/2))
  end do
  write(*,93) "Standard errors of MLEs:"
  write(*,90) se
  stop
90 FORMAT(' ',6g13.6)
93 FORMAT(/' ',a,g15.8)
  end program mle


subroutine obj_f(npar, param, nf, dev, inputs, Xy, ufparm)
  use sn
  implicit none
  INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(15, 60)
  ! real(wp), parameter :: pi = 3.141592653
  real(wp), parameter :: max_gamma1 = 0.9952719
  real(dp), intent(in) :: param(:)
  real(dp), intent(out) :: dev
  real(dp), intent(in out) ::  Xy(:), ufparm
  integer, intent(in) :: npar, nf, inputs(:)
  real(wp) :: cp(size(param))
  logical :: trace
  integer ::  n,p
  real(wp) :: penalty
  real(wp), dimension(inputs(1),inputs(2)) :: X
  real(wp), dimension(inputs(1)) :: y
  n=inputs(1)
  p=inputs(2)
  if(npar /= (p+2)) stop ! "wrong parameters"
  X(1:n,1:p)=reshape(Xy(1:n*p),(/n,p/))
  y=Xy((n*p+1):n*(p+1))
  trace=(inputs(3)==1)
  cp=param
  penalty=0.0
  if(cp(p+1) <= zero) penalty=penalty-(one+cp(p+1))*10.0e20
  if(abs(cp(p+2)) > max_gamma1) penalty=penalty+abs(cp(p+2))*10.e20
  if(penalty>0) then
    dev = penalty
  else
    dev = sn_dev(cp, X, y, trace)
  end if
  if(npar<0) ufparm=nf ! ELF90 requirement
  return
end subroutine obj_f

subroutine obj_gh(npar, param, nf, dev_g, dev_h, inputs, Xy, ufparm)
  use sn
  implicit none
  INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(15, 60)
  ! real(wp), parameter :: pi = 3.141592653
  real(wp), parameter :: max_gamma1 = 0.9952719
  real(dp), intent(in) :: param(:)
  real(dp), intent(out) :: dev_g(:), dev_h(:)
  real(dp), intent(in out) ::  Xy(:), ufparm
  integer, intent(in) :: npar, nf, inputs(:)
  real(wp), dimension(size(param)) :: cp, grad
  logical :: trace
  integer ::  n,p,i,j
  real(wp) :: penalty, info(npar,npar)
  real(wp), dimension(inputs(1),inputs(2)) :: X
  real(wp), dimension(inputs(1)) :: y
  n=inputs(1)
  p=inputs(2)
  if(npar /= (p+2) .OR. size(grad) /= npar) stop !"wrong parameters"
  X(1:n,1:p)=reshape(Xy(1:n*p),(/n,p/))
  y=Xy((n*p+1):n*(p+1))
  trace=(inputs(3)==1)
  cp=param
  penalty=0.0
  if(cp(p+1) <= 0.0) penalty=penalty-(1.0+cp(p+1))*10.0e20
  if(abs(cp(p+2)) > max_gamma1) penalty=penalty+abs(cp(p+2))*10.e20
  ! write(*,'(1x,A,(5g17.8))') "param:", param
  if(penalty>0) then
    dev_g = 0.0*param
    dev_h(1:(npar+1)*npar/2)=zero
  else
    call sn_dev_gh(cp, X, y, grad, info, trace)
    dev_g(1:npar)=grad
    do i=1,npar
      do j=1,i
        dev_h(i*(i-1)/2+j)=2.0_dp*info(i,j)
      end do
    end do
  end if
  if(npar<0) ufparm=nf+info(1,1) ! fake ELF90 
  return
end subroutine obj_gh

subroutine prompt_int(message,value)
  implicit none
  character (LEN=*), INTENT(IN) ::message
  INTEGER, INTENT(OUT) :: value
  write(*,'(a)',ADVANCE='no') message
  READ(*,*) value
  return
end subroutine prompt_int



