! Author(except for ALNORM): 
! Adelchi Azzalini, Universit\`a di Padova, Italia
! December 1998,  trivial modification in 2008

module normal

IMPLICIT NONE
INTEGER, PARAMETER  :: wp = SELECTED_REAL_KIND(15, 60)
real(wp), parameter :: zero = 0.0_wp, one = 1.0_wp, half = 0.5_wp, two=2.0_wp
real(wp), parameter :: pi =  3.14159265358979_wp
real(wp), parameter :: root2pi = 2.506628274631_wp
real(wp), parameter :: log_root2pi= (-0.918938533204673)

contains
function dnorm(x)
  implicit none
  real(wp), intent(in), dimension(:) :: x
  real(wp), dimension(size(x)) :: dnorm
  
  dnorm(1:size(x))=exp(-0.5_wp*x*x)/root2pi
  return
end function dnorm

function pnorm(x)  
  implicit none
  real(wp), intent(in), dimension(:) :: x
  real(wp), dimension(size(x)) :: pnorm
  integer :: i
  do i=1,size(x)
    pnorm(i)=alnorm(x(i),.FALSE.)
  end do
  return
end function pnorm

FUNCTION alnorm(x, upper) RESULT(fn_val)

!  Algorithm AS66 Applied Statistics (1973) vol.22, no.3

!  Evaluates the tail area of the standardised normal curve
!  from x to infinity if upper is .true. or
!  from minus infinity to x if upper is .false.

! ELF90-compatible version by Alan Miller
! Latest revision - 29 November 1997

IMPLICIT NONE
real(wp), INTENT(IN) :: x
LOGICAL, INTENT(IN)  :: upper
real(wp)             :: fn_val

! Local variables  zero = 0.0_wp, one = 1.0_wp, half = 0.5_wp, &
real(wp), PARAMETER ::  con = 1.28_wp
real(wp) :: z, y
LOGICAL   :: up

!*** machine dependent constants
real(wp), PARAMETER :: ltone = 7.0_wp, utzero = 18.66_wp

real(wp), PARAMETER ::  p = 0.398942280444_wp, q = 0.39990348504_wp,   &
                        r = 0.398942280385_wp, a1 = 5.75885480458_wp,  &
                        a2 = 2.62433121679_wp, a3 = 5.92885724438_wp,  &
                        b1 = -29.8213557807_wp, b2 = 48.6959930692_wp, &
                        c1 = -3.8052E-8_wp, c2 = 3.98064794E-4_wp,     &
                        c3 = -0.151679116635_wp, c4 = 4.8385912808_wp, &
                        c5 = 0.742380924027_wp, c6 = 3.99019417011_wp, &
                        d1 = 1.00000615302_wp, d2 = 1.98615381364_wp,  &
                        d3 = 5.29330324926_wp, d4 = -15.1508972451_wp, &
                        d5 = 30.789933034_wp

up = upper
z = x
IF(z >=  zero) GO TO 10
up = .NOT. up
z = -z
10 IF(z <= ltone .OR. up .AND. z <= utzero) GO TO 20
fn_val = zero
GO TO 40
20 y = half*z*z
IF(z > con) GO TO 30

fn_val = half - z*(p-q*y/(y+a1+b1/(y+a2+b2/(y+a3))))
GO TO 40
30 fn_val = r*EXP(-y)/(z+c1+d1/(z+c2+d2/(z+c3+d3/(z+c4+d4/(z+c5+d5/(z+c6))))))
40 IF(.NOT. up) fn_val = one - fn_val

RETURN
END FUNCTION alnorm

end module normal


!----------------------------------------------------------------------

module sn

use normal
implicit none
! real(wp), parameter :: zero = 0.0_wp, one = 1.0_wp,         &
!                       two=2.0_wp,  half = 0.5_wp,           &
!                       pi =  3.14159265358979_wp,            &
!                       log_root2pi= (-0.918938533204673)
character(len=*), parameter :: form="(1x,a,(1x,5g15.7))"
contains

recursive function zeta(k,x)  result(y)
! k integer \in (0,4). 
  implicit none
  integer, intent(in) :: k
  real(wp), intent(in), dimension(:) :: x
  real(wp), dimension(size(x)) :: y
  logical, dimension(size(x)) :: ok
  ok = (-35.0_wp < x)
  select case (k)
  case(0)
    where(ok)
      y=  log(two*pnorm(x))
    elsewhere
      y = log_root2pi-half*x*x-log(x)
    end where
  case(1)
      where(ok)
        y =dnorm(x)/pnorm(x)        
      elsewhere
        y  = (-x)*(1+1/x**2)
      end where
  case (2)
      y=(-zeta(1,x)*(x+zeta(1,x)))
  case(3) 
      y=(-zeta(2,x)*(x+zeta(1,x))-zeta(1,x)*(1+zeta(2,x)))
  case (4)
      y=(-zeta(3,x)*(x+2*zeta(1,x))-2*zeta(2,x)*(1+zeta(2,x)))
  case default
      call error("zeta(k,x) called with k not in 0..4")
  end select
  return
end function zeta

function cp_to_dp(param)
  ! converts centred parameters cp=(mu,sigma,gamma1)
  ! to direct parameters dp=(xi,omega,lambda)
  ! Note:  mu can be m-dimenstional, the other musts be scalars
  implicit none
  real(wp), intent(in), dimension(:) :: param
  real(wp), dimension(size(param)-2) :: location
  real(wp), dimension(size(param)) :: cp_to_dp
  real(wp) ::  A, b, lambda, delta, E_Z, sd_Z, scale, gamma1
  integer :: m
  b = sqrt(two/pi)
  m = size(param)-2
  gamma1 = param(m+2)
  if(abs(gamma1)>0.9952719) call ERROR("cp_to_dp: abs(gamma1)>0.9952719")
  A = sign((abs(two*gamma1/(4.0_wp-pi)))**(one/3.0_wp), gamma1)
  delta = A/(b*sqrt(one+A**2))
  lambda = delta/sqrt(one-delta**2)
  E_Z = b*delta
  sd_Z = sqrt(one-E_Z**2)
  location    = param(1:m)
  location(1) = param(1)-param(m+1)*E_Z/sd_Z
  scale = param(m+1)/sd_Z
  cp_to_dp(1:m) = location
  cp_to_dp(m+1) = scale
  cp_to_dp(m+2) = lambda
  return
  end function cp_to_dp

function gamma1_to_lambda(gamma1) 
  implicit none
  real(wp), parameter :: pi = 3.14159265358979_wp
  real(wp), parameter :: max_gamma1 = 0.995271746431_wp
  real(wp), intent(in), dimension(:) :: gamma1
  real(wp), dimension(size(gamma1)) :: a, lambda, delta, gamma1_to_lambda
  ! max_gamma1 = half*(4-pi)*(two/(pi-two))**1.5
  ! na = (abs(gamma1)>max_gamma1)
  ! if(any(na)) warning("NAs generated") 
  ! gamma1=replace(gamma1,na,NA)
  if(any(abs(gamma1)>max_gamma1))  &
                  CALL error("gamma1_to_lambda: gamma1 not feasible")
  a    = sign((two*abs(gamma1)/(4.0-pi))**(one/3.0_wp), gamma1)
  delta= sqrt(pi/two)*a/sqrt(1+a**2)
  lambda=delta/sqrt(one-delta**2)
  gamma1_to_lambda=lambda
  return
  end function gamma1_to_lambda

function sn_dev (cp, X, y, trace)
! -2*logL for centred parameters
  implicit none
  real(wp), parameter :: root2pi = 2.506628274631_wp
  real(wp), intent(in), dimension (:,:) :: X
  real(wp), intent(in), dimension (:) :: y
  real(wp), intent(in), dimension (:) :: cp
  logical, intent(in), optional :: trace
  real(wp), dimension (size(cp)) :: dp
  real(wp), dimension (size(y)) :: location
  real(wp), dimension (size(y)) :: z
  real(wp) :: sn_dev, nlogL, scale
  integer :: m, n
  n = size(y)
  m = size(X,2)
  dp = cp_to_dp(cp)
  location = matmul(X, dp(1:m))
  scale=dp(m+1)
  ! AVOID: logL = sum(log(dsn(y,location,dp(m+1),dp(m+2))))
  z=(y-location)/scale
  nlogL = n*log(root2pi*scale) + half*sum(z*z) - sum(zeta(0,dp(m+2)*z))
  if(trace)  write(*,form)  "sn_dev\(cp,logL): ", cp, -nlogL
  sn_dev=2*nlogL
  return
  end function sn_dev

subroutine sn_dev_gh(cp, X, y, gradient, info, trace)
  ! computes gradient and hessian of dev=-2*logL for centred parameters
  ! (and observed information matrix);
  ! returns list with hessian in the format required by nlminb
  implicit none
  real(wp), intent(in) :: cp(:), X(:,:), y(:)
  real(wp), intent(out) :: gradient(:),  info(:,:)
  logical, intent(in), optional :: trace
  real(wp) :: sigma, gamma1, lambda, E_Z, s_Z, omega, DE_Z, Ds_Z, DDE_Z,      &
        DDs_Z, Dg_Dl, Vtmp(size(X,1)), Xtmp(size(X,1),size(X,2))
  real(wp), allocatable :: beta(:), mu(:), d(:), r(:), z(:), p1(:), p2(:),    &
        w(:), Dz(:), DDz(:), score(:)
  integer :: i, n, p, np
  p  = size(X,2)
  n  = size(X,1)
  np = p+2
  allocate (beta(p), mu(n), d(n), r(n), z(n), p1(n), p2(n), w(n), Dz(n),  &
        DDz(n), score(size(cp)))
  beta = cp(1:p)
  sigma = cp(p+1)
  gamma1 = cp(p+2)
  gradient(1:1) = gamma1_to_lambda((/gamma1/))
  lambda=gradient(1)
  mu = matmul(X, beta)
  d  = y-mu
  r  = d/sigma
  E_Z= lambda*sqrt(two/(pi*(one+lambda**2)))
  s_Z= sqrt(one-E_Z**2)
  z  = E_Z+s_Z*r
  p1 = zeta(1,lambda*z)
  p2 = zeta(2,lambda*z)
  omega= sigma/s_Z
  w    = lambda*p1-E_Z
  DE_Z = sqrt(two/pi)/(one+lambda**2)**1.5
  Ds_Z = (-E_Z/s_Z)*DE_Z
  Dz   = DE_Z + r*Ds_Z
  DDE_Z= (-3.0)*E_Z/(1+lambda**2)**2
  DDs_Z= -((DE_Z*s_Z-E_Z*Ds_Z)*DE_Z/s_Z**2+E_Z*DDE_Z/s_Z)
  DDz  = DDE_Z + r*DDs_Z
  score(1:p) = omega**(-2)*matmul(transpose(X),(y-mu-omega*w))
  score(p+1) = (-n)/sigma+s_Z*dot_product(d,z-p1*lambda)/sigma**2
  score(p+2) = n*Ds_Z/s_Z-dot_product(z,Dz)+dot_product(p1,z+lambda*Dz)
  Dg_Dl = 1.5*(4.0-pi)*E_Z**2*(DE_Z*s_Z-E_Z*Ds_Z)/s_Z**4
  score(p+2) = score(p+2)/Dg_Dl  ! convert deriv wrt lamda to gamma1
  !info(1:p,1:p) = omega**(-2) * matmul(transpose(X),(one-lambda**2*p2)*X)
  do i=1,n
    Vtmp(i)=1.0_wp-lambda**2*p2(i)
  end do
  do i=1,n
    Xtmp(i,1:p)=X(i,1:p)*Vtmp(i)
  end do
  info(1:p,1:p) = omega**(-2) * matmul(transpose(X),Xtmp)
  info(p+1,1:p) = s_Z*matmul(transpose(X), (z-lambda*p1)+d*Vtmp/omega)/sigma**2
  info(1:p,p+1) = info(p+1,1:p)
  info(p+1,p+1) = (-n)/sigma**2+2*s_Z*sum(d*(z-lambda*p1))/sigma**3 +                &
           s_Z**2*sum(d*Vtmp*d)/sigma**4
  info(p+2,1:p) = matmul( transpose(X),                                              &
           (-2*Ds_Z*d/omega+Ds_Z*w+s_Z*(p1+lambda*p2*(z+lambda*Dz)-DE_Z)))/sigma
  info(1:p,p+2) = info(p+2,1:p)
  info(p+2,p+1) = -sum (d*(Ds_Z*(z-lambda*p1)+                                       &
          s_Z*(Dz-p1-p2*lambda*(z+lambda*Dz))))/sigma**2
  info(p+1,p+2) = info(p+2,p+1)
  info(p+2,p+2) = n*(-DDs_Z*s_Z+Ds_Z**2)/s_Z**2+sum(Dz**2+z*DDz)-                    &
           sum(p2*(z+lambda*Dz)**2)- sum(p1*(2*Dz+lambda*DDz))
  info(np,1:np) = info(np,1:np)/Dg_Dl ! convert info wrt lamda to gamma1
  info(1:np,np) = info(1:np,np)/Dg_Dl
  !if(trace) {cat("grad_hessian: gradient="); print(-2*score)}
  !list(gradient=-2*score, hessian=2*info(!lower_tri(info)), info=info)
  gradient= -two*score 
  if(trace) write(*,form)"sn_dev_gh\gradient: ",gradient
  return
end subroutine sn_dev_gh

subroutine error(message)
  character (LEN=*), INTENT(IN) ::message
  WRITE(*,*) message
  stop
end subroutine error

end module sn
